#ifndef _PT_Buffer_H
#define _PT_Buffer_H

#include "EARTH_PT.h"
#include "Prefix.h"
#include "Memory.h"
#include "Memory_Block.h"
#include "XC3S.h"
#include <vector>
#include <Windows.h>

namespace EARTH {
namespace Memory { class _; class Block;  }
namespace PT {

	class Buffer {
	public:
		Buffer(uint card, Memory::_ *memory, XC3S::_ *xc3s) : mCard(card), mMemory(memory), mXC3S(xc3s), mImageSize(511)
		{
			mBufferInfo.VirtualSize = 0;
			mBufferInfo.VirtualCount = 0;
			mBufferInfo.LockSize = 0;
		}

		~Buffer()
		{
		}

		status SetBufferInfo(const Device::BufferInfo *bufferInfo)
		{
			status status;

			if (bufferInfo) {
				status = Alloc(bufferInfo);
			} else {
				status = Free();
			}

			return status;
		}

		status GetBufferInfo(Device::BufferInfo *bufferInfo) const
		{
			if (bufferInfo == NULL) return STATUS_INVALID_PARAM_ERROR;

			*bufferInfo = mBufferInfo;

			return STATUS_OK;
		}

		status GetBufferPtr(uint index, void **ptr_) const
		{
			if (ptr_ == NULL) return STATUS_INVALID_PARAM_ERROR;

			if (mBufferInfo.VirtualCount <= index) return STATUS_INVALID_PARAM_ERROR;

			const std::vector<void *> &bufferVector = mBufferVector[0];
			if (bufferVector.size() <= index) return STATUS_INTERNAL_ERROR;

			void *ptr = bufferVector.at(index);
			if (ptr == NULL) return STATUS_INTERNAL_ERROR;
			*ptr_ = ptr;
			
			return STATUS_OK;
		}

		status GetStartAddress(uint *startAddress) const
		{
			const std::vector<Memory::Block *> &blockVector = mBlockVector[1];
			if (blockVector.size() == 0) return STATUS_INTERNAL_ERROR;

			Memory::Block *block = blockVector.at(0);

			uint infoCount;
			const Memory::Block::PhysicalAddressInfo *infoTable;
			block->GetPhysicalAddress(&infoCount, &infoTable);
			if (infoCount == 0) return STATUS_INTERNAL_ERROR;

			const Memory::Block::PhysicalAddressInfo *info = &infoTable[0];
			ulong physicalAddress = info->PhysicalAddress;
			uint physicalAddress32;
			status status = ConvertAddress(physicalAddress, &physicalAddress32);
			if (status) return status;

			_ASSERTE(startAddress);
			*startAddress = physicalAddress32;

			return status;
		}

	protected:
		Memory::_ *const mMemory;
		const uint mCard;
		XC3S::_ *const mXC3S;

		const uint mImageSize;
		Device::BufferInfo mBufferInfo;

		std::vector<void *> mBufferVector[2];
		std::vector<Memory::Block *> mBlockVector[2];

		status Alloc(const Device::BufferInfo *bufferInfo)
		{
			if (bufferInfo == NULL) return STATUS_INVALID_PARAM_ERROR;
			if ((bufferInfo->VirtualSize == 0) || (bufferInfo->VirtualCount == 0) || (bufferInfo->LockSize == 0)) return STATUS_INVALID_PARAM_ERROR;
			if (bufferInfo->VirtualSize % bufferInfo->LockSize) return STATUS_INVALID_PARAM_ERROR;

			if (mBufferInfo.VirtualSize) return STATUS_BUFFER_ALREADY_ALLOCATED_ERROR;
			
			status status;

			status = AllocData(bufferInfo);
			if (status) { Free(); return status; }

			status = AllocAddress(bufferInfo);
			if (status) { Free(); return status; }

			std::vector<Memory::Block *> *blockVector;

			for (uint i=0; i<sizeof(mBlockVector)/sizeof(*mBlockVector); i++) {
				blockVector = &mBlockVector[i];
				for (uint j=0; j<blockVector->size(); j++) {
					Memory::Block *block = blockVector->at(j);
					status = block->Lock((i == 0) ? Memory::Block::DIRECTION_FROM_DEVICE : Memory::Block::DIRECTION_TO_DEVICE);
					if (status) { Free(); return status; }
				}
			}

			status = LinkData();
			if (status) { Free(); return status; }

			status = LinkAddress();
			if (status) { Free(); return status; }

			mBufferInfo = *bufferInfo;

			return status;
		}

		status Free()
		{
			// Unlock
			for (uint i=0; i<sizeof(mBlockVector)/sizeof(*mBlockVector); i++) {
				std::vector<Memory::Block *> *vector = &mBlockVector[i];
				_ASSERTE(vector);
				for (uint j=0; j<vector->size(); j++) {
					Memory::Block *block = vector->at(j);
					_ASSERTE(block);
					if (block->IsLocked()) {
						block->Unlock();
					}
					delete_(block);
				}
				vector->clear();
			}

			// Free
			for (uint i=0; i<sizeof(mBufferVector)/sizeof(*mBufferVector); i++) {
				std::vector<void *> *vector = &mBufferVector[i];
				_ASSERTE(vector);
				for (uint j=0; j<vector->size(); j++) {
					void *buffer = vector->at(j);
					_ASSERTE(buffer);
					Memory::Free(buffer);
					buffer = NULL;
				}
				vector->clear();
			}

			mBufferInfo.VirtualSize = 0;
			mBufferInfo.VirtualCount = 0;
			mBufferInfo.LockSize = 0;

			return STATUS_OK;
		}

		status AllocData(const Device::BufferInfo *bufferInfo)
		{
			_ASSERTE(bufferInfo);

			std::vector<void *> &bufferVector = mBufferVector[0];
			std::vector<Memory::Block *> &blockVector = mBlockVector[0];
			
			_ASSERTE((bufferInfo->VirtualSize % bufferInfo->LockSize) == 0);
			uint lockCount = bufferInfo->VirtualSize / bufferInfo->LockSize;

			_ASSERTE(bufferVector.size() == 0);
			for (uint i=0; i<bufferInfo->VirtualCount; i++) {
				//wprintf(L"AllocData() buffer: %d y[W\n", mImageSize*mVirtualSize);
				void *buffer = Memory::Alloc(Memory::SIZE * mImageSize * bufferInfo->VirtualSize);
				if (buffer == NULL) return STATUS_VIRTUAL_ALLOC_ERROR;
				bufferVector.push_back(buffer);

				for (uint j=0; j<lockCount; j++) {
					//wprintf(L"AllocData() block: %d y[W\n", mImageSize*mLockSize);
					Memory::Block *block = mMemory->NewBlock(mCard, Memory::SIZE * mImageSize * bufferInfo->LockSize * j, Memory::SIZE * mImageSize * bufferInfo->LockSize);
					if (block == NULL) return STATUS_OUT_OF_MEMORY_ERROR;
					block->SetBasePtr(buffer);
					blockVector.push_back(block);
				}
			}

			return STATUS_OK;
		}

		status AllocAddress(const Device::BufferInfo *bufferInfo)
		{
			_ASSERTE(bufferInfo);

			std::vector<void *> &bufferVector = mBufferVector[1];
			std::vector<Memory::Block *> &blockVector = mBlockVector[1];

			_ASSERTE(bufferVector.size() == 0);
			uint pageCount = bufferInfo->VirtualSize * bufferInfo->VirtualCount;
			while (pageCount) {
				uint pageCount2 = (pageCount <= 512) ? pageCount : 512;
				//wprintf(L"AllocAddress(): %d y[W\n", pageCount2);
				void *buffer = Memory::Alloc(Memory::SIZE*pageCount2);
				if (buffer == NULL) return STATUS_VIRTUAL_ALLOC_ERROR;
				bufferVector.push_back(buffer);

				Memory::Block *block = mMemory->NewBlock(mCard, 0, Memory::SIZE * pageCount2);
				if (block == NULL) return STATUS_OUT_OF_MEMORY_ERROR;
				block->SetBasePtr(buffer);
				blockVector.push_back(block);

				pageCount -= pageCount2;
			}

			return STATUS_OK;
		}

		status LinkData()
		{
			uint dstWordCount = 0;
			uint dstIndex = 0;
			uint *dstPtr = NULL;
			uint dstPageCount = 0;

			for (uint i=0; i<mBlockVector[0].size(); i++) {
				Memory::Block *srcBlock = mBlockVector[0].at(i);

				uint infoCount;
				const Memory::Block::PhysicalAddressInfo *infoTable;
				srcBlock->GetPhysicalAddress(&infoCount, &infoTable);

				for (uint j=0; j<infoCount; j++) {
					const Memory::Block::PhysicalAddressInfo *info = &infoTable[j];
					ulong physicalAddress = info->PhysicalAddress;
					_ASSERTE((info->Size % Memory::SIZE) == 0);
					uint pageCount = info->Size / Memory::SIZE;
					for (uint k=0; k<pageCount; k++) {
						if (dstWordCount == 0) {
							if (dstPageCount == 0) {
								Memory::Block *dstBlock = mBlockVector[1].at(dstIndex++);
								dstPtr = static_cast<uint *>(dstBlock->Ptr());
								dstPtr += 1;
								dstPageCount = dstBlock->Size() / Memory::SIZE;
								dstWordCount = mImageSize;
							} else {
								dstPageCount--;
								dstWordCount = mImageSize;
								dstPtr += Memory::SIZE/sizeof(uint) - (1 + mImageSize);
								dstPtr += 1;
							}
						}

						uint physicalAddress32;
						status status = ConvertAddress(physicalAddress, &physicalAddress32);
						if (status) return status;

						_ASSERTE(dstPtr);
						*dstPtr++ = physicalAddress32;
						physicalAddress += Memory::SIZE;
						dstWordCount--;
					}
				}
			}

			return STATUS_OK;
		}

		status LinkAddress()
		{
			bool first = true;
			ulong physicalAddress0 = 0;
			uint dstCount = 0;
			uint dstIndex = 0;
			uint *dstPtr = NULL;

			for (uint i=0; i<mBlockVector[1].size(); i++) {
				Memory::Block *srcBlock = mBlockVector[1].at(i);

				uint infoCount;
				const Memory::Block::PhysicalAddressInfo *infoTable;
				srcBlock->GetPhysicalAddress(&infoCount, &infoTable);

				for (uint j=0; j<infoCount; j++) {
					const Memory::Block::PhysicalAddressInfo *info = &infoTable[j];
					ulong physicalAddress = info->PhysicalAddress;
					if (first) {
						physicalAddress0 = physicalAddress;
					}
					_ASSERTE((info->Size % Memory::SIZE) == 0);
					uint srcCount = info->Size / Memory::SIZE;
					for (uint k=0; k<srcCount; k++) {
						if (dstCount == 0) {
							Memory::Block *dstBlock = mBlockVector[1].at(dstIndex++);
							dstPtr = static_cast<uint *>(dstBlock->Ptr());
							dstCount = dstBlock->Size() / Memory::SIZE;
						}

						if (first) {
							first = false;
						} else {
							uint physicalAddress32;
							status status = ConvertAddress(physicalAddress, &physicalAddress32);
							if (status) return status;

							_ASSERTE(dstPtr);
							*dstPtr = physicalAddress32;
							dstPtr += Memory::SIZE / sizeof(*dstPtr);
							dstCount--;
						}
						physicalAddress += Memory::SIZE;
					}
				}
			}

			if (dstPtr) {
				uint physicalAddress32;
				status status = ConvertAddress(physicalAddress0, &physicalAddress32);
				if (status) return status;
				*dstPtr = physicalAddress32;
			}

			return STATUS_OK;
		}

		status ConvertAddress(ulong address64, uint *address32) const
		{
			_ASSERTE((address64 % Memory::SIZE) == 0);

			address64 >>= 10+2;

			if (address64 >> 31) {
				return STATUS_DMA_ADDRESS_ERROR;
			}

			uint address = static_cast<uint>(address64);

			bool useDAC = (address >> 20) || false/*mAlwaysUseDAC*/;
			if (useDAC) {
				address |= BIT_1(1) << 31;

				#if defined(_DEBUG)
				static bool gFirst = true;
				if (gFirst) {
					::MessageBox(NULL, L"useDAC", L"DMA", MB_OK);
					gFirst = false;
				}
				#endif
			}

			_ASSERTE(address32);
			*address32 = address;

			return STATUS_OK;
		}
	};
}
}

#endif
