#ifndef _PT_Device_H
#define _PT_Device_H

#include "EARTH_PT.h"
#include "Prefix.h"
#include "XC3S.h"
#include "TC90512.h"
#include "TDA6651.h"
#include "QM1D1B0004.h"
#include "PT_Buffer.h"
#include "PT_Device2.h"

namespace EARTH {
namespace Memory { class _; class Block;  }
namespace PCI    { class _; class Device; }
namespace PT {

	class Bus_;
	class Device2_;

	class Device_ : public Device {
	public:
		Device_(void (*busFunction)(Device_ *, void *), void *busArgument, PCI::_ *pci, Memory::_ *memory, const Bus::DeviceInfo *deviceInfo)
			: mBusFunction(busFunction), mBusArgument(busArgument), mPCI(pci), mMemory(memory), mDeviceInfo(*deviceInfo)
		{
			mDevice = NULL;
			mXC3S = NULL;
			mAT45DB041D = NULL;

			mDevice2 = NULL;
			mBuffer = NULL;

			mTransferEnable = false;

			_ASSERTE(mDevice2 == NULL);
			mDevice2 = new Device2_();

			for (uint i=0; i<sizeof(mTuner)/sizeof(*mTuner); i++) {
				Tuner *tuner = &mTuner[i];
				_ASSERTE(tuner);
				tuner->TC90512 = NULL;
				tuner->TDA6651 = NULL;
				tuner->QM1D1B0004 = NULL;
				for (uint j=0; j<2; j++) {
					tuner->Sleep           [j] = true;
					tuner->FrequencyChannel[j] = 0;
					tuner->FrequencyOffset [j] = 0;
				}
			}
		}

		~Device_()
		{
			delete_(mDevice2);

			mBusFunction(this, mBusArgument);
		}

		virtual status Delete()
		{
			if (mDevice) return STATUS_DEVICE_MUST_BE_CLOSED_ERROR;

			delete this;

			return STATUS_OK;
		}

		Device2 *Device2() const
		{
			return mDevice2;
		}

		virtual status Open()
		{
			if (mDevice) return STATUS_DEVICE_IS_ALREADY_OPEN_ERROR;

			status status;

			PCI::Target target;
			target.Bus      = mDeviceInfo.Bus;
			target.Slot     = mDeviceInfo.Slot;
			target.Function = mDeviceInfo.Function;

			status = mPCI->NewDevice(&target, &mDevice);
			if (status) return status;

			byte configRevision;
			status = mDevice->ReadConfig(0x08, &configRevision, sizeof(configRevision));
			if (status) return status;

			if (configRevision != 0x01) return STATUS_CONFIG_REVISION_ERROR;

			XC3S::_::Argument argument;
			argument.Pci    = mDevice;
			argument.Memory = mMemory;

			_ASSERTE(mXC3S == NULL);
			mXC3S = static_cast<XC3S ::_ *>(new XC3S::_(&argument));

			_ASSERTE(mDevice2);
			mDevice2->SetXC3S(mXC3S);

			status = mXC3S->Config()->CheckBus();
			if (status) return status;

//			uint address = 0xfebff000;	// GF6150
			uint address = 0xfeaff000;	// P5B
//			uint address = 0xe8125000;	// SiS
			status = SetConfig(&address);
			if (status) return status;

			uint version;
			mXC3S->Config()->GetVersion(&version);
			
			XC3S::I2C *i2c = mXC3S->I2C();
			_ASSERTE(i2c);

			for (uint i=0; i<sizeof(mTuner)/sizeof(*mTuner); i++) {
				Tuner *tuner = &mTuner[i];
				_ASSERTE(tuner);

				_ASSERTE(tuner->TC90512 == NULL);
				tuner->TC90512 = new TC90512(i2c, i^1, 20.48*4);

				_ASSERTE(tuner->TDA6651 == NULL);
				tuner->TDA6651 = new TDA6651(tuner->TC90512);

				_ASSERTE(tuner->QM1D1B0004 == NULL);
				tuner->QM1D1B0004 = new QM1D1B0004(tuner->TC90512);
			}

			_ASSERTE(mAT45DB041D == NULL);
			mAT45DB041D = new AT45DB041D(mXC3S->Register());

			_ASSERTE(mDevice2);
			mDevice2->SetAT45DB041D(mAT45DB041D);
			mDevice2->SetVersion(version);

			bool currentVersion = (version == 0x00000001);
			if (currentVersion == false) return STATUS_FPGA_VERSION_ERROR;

			mXC3S->SetVersion(version);

			uint dataWhenClose = mXC3S->Initiator()->DataWhenClose();

			status = mXC3S->Clock()->SetReset(dataWhenClose);
			if (status) return status;

			status = mXC3S->PowerReset()->SetLnbPowerWhenClose(Device::LNB_POWER_OFF);
			if (status) return status;

			mXC3S->Register()->EnableWrite();
			mXC3S->Register()->EnableRead ();

			i2c->Init();

			#ifdef _DEBUG
			mDevice2->UnlockDebug();
			#endif
	
			status = mXC3S->Protect()->Run();
			if (status) return status;

			status = mXC3S->Clock()->EnableRAM();
			if (status) return status;

			for (uint i=0; i<2; i++) {
				for (uint j=0; j<2; j++) {
					ISDB isdb = static_cast<ISDB>(j);
					mXC3S->Initiator()->SetStreamEnable(i, isdb, false);
					mXC3S->Initiator()->SetStreamEnable(i, isdb, 0);
				}
			}

			_ASSERTE(mBuffer == NULL);
			mBuffer = new class Buffer(mDevice->ID(), mMemory, mXC3S);
			_ASSERTE(mBuffer);

			return status;
		}

		virtual status Close()
		{
			status status;

			status = Status();
			if (status) return status;

			if (mTransferEnable) {
				SetTransferEnable(false);
			}

			delete_(mBuffer);

			delete_(mAT45DB041D);

			_ASSERTE(mDevice2);
			mDevice2->SetAT45DB041D(mAT45DB041D);

			for (uint i=0; i<sizeof(mTuner)/sizeof(*mTuner); i++) {
				Tuner *tuner = &mTuner[i];
				_ASSERTE(tuner);
				delete_(tuner->TC90512);
				delete_(tuner->TDA6651);
				delete_(tuner->QM1D1B0004);

				for (uint j=0; j<2; j++) {
					tuner->Sleep           [j] = true;
					tuner->FrequencyChannel[j] = 0;
					tuner->FrequencyOffset [j] = 0;
				}				
			}

			delete_(mXC3S);

			_ASSERTE(mDevice2);
			mDevice2->SetXC3S(mXC3S);

			if (mDevice) {
				mDevice->Delete();
				mDevice = NULL;
			}

			return status;
		}

		virtual status GetPciClockCounter(uint *counter)
		{
			status status;

			status = Status();
			if (status) return status;

			if (counter == NULL) return STATUS_INVALID_PARAM_ERROR;

			mXC3S->Counter()->GetPci(counter);
	
			return status;
		}

		virtual status SetPciLatencyTimer(byte latencyTimer)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->Config()->SetLatencyTimer(latencyTimer);

			return status;
		}

		virtual status GetPciLatencyTimer(byte *latencyTimer) const
		{
			status status;

			status = Status();
			if (status) return status;

			if (latencyTimer == NULL) return STATUS_INVALID_PARAM_ERROR;

			status = mXC3S->Config()->GetLatencyTimer(latencyTimer);

			return status;
		}

		virtual status SetLnbPower(LnbPower lnbPower)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->PowerReset()->SetLnbPower(lnbPower);
	
			return status;
		}

		virtual status GetLnbPower(LnbPower *lnbPower) const
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->PowerReset()->GetLnbPower(lnbPower);

			return status;
		}

		virtual status SetLnbPowerWhenClose(LnbPower lnbPower)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->PowerReset()->SetLnbPowerWhenClose(lnbPower);

			return status;
		}

		virtual status GetLnbPowerWhenClose(LnbPower *lnbPower) const
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->PowerReset()->GetLnbPowerWhenClose(lnbPower);

			return status;
		}

		virtual status SetTunerPowerReset(TunerPowerReset tunerPowerReset)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->PowerReset()->SetTunerPowerReset(tunerPowerReset);

			return status;
		}

		virtual status GetTunerPowerReset(TunerPowerReset *tunerPowerReset) const
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->PowerReset()->GetTunerPowerReset(tunerPowerReset);

			return status;
		}

		virtual status InitTuner(uint tunerIndex)
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];
			_ASSERTE(tuner);

			status = tuner->TC90512->Init();
			if (status) return status;

			tuner->QM1D1B0004->Write(0);
			tuner->TDA6651->Write(0, TDA6651::Mode_Normal);

			for (uint j=0; j<2; j++) {
				tuner->Sleep           [j] = true;
				tuner->FrequencyChannel[j] = 0;
				tuner->FrequencyOffset [j] = 0;
			}

			return status;
		}

		virtual status SetTunerSleep(uint tunerIndex, ISDB isdb, bool sleep)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];
			_ASSERTE(tuner);

			switch (isdb) {
				case ISDB_S:
				{
					if (sleep) {
						tuner->QM1D1B0004->Write(0);
					}
					tuner->TC90512->PSK_Write_waitm(sleep ? 1 : 0);
					break;
				}
				case ISDB_T:
				{
					if (sleep) {
						tuner->TDA6651->Write(0, TDA6651::Mode_Normal);
					}
					tuner->TC90512->OFDM_Write_slptim(sleep ? 1 : 0);
					break;
				}
			}
	
			tuner->Sleep[isdb] = sleep;

			return status;
		}

		virtual status GetTunerSleep(uint tunerIndex, ISDB isdb, bool *sleep) const
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			if (sleep == NULL) return STATUS_INVALID_PARAM_ERROR;

			const Tuner *tuner = &mTuner[tunerIndex];
			_ASSERTE(tuner);
			*sleep = tuner->Sleep[isdb];

			return STATUS_OK;
		}

		virtual status SetFrequency(uint tunerIndex, ISDB isdb, uint channel, int offset)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					bool bs;
					uint number, frequency;

					GetSatelliteFrequency(channel, &bs, &number, &frequency);

					tuner->QM1D1B0004->Write((frequency + 50) / 100 + offset);

					bool locked = false;
					for (uint i=0; i<50; i++) {
						tuner->QM1D1B0004->Read(&locked);
						if (locked) break;
						OS::Thread::Sleep((1)*1000*1000);
					}
					if (locked == false) return STATUS_PLL_LOCK_TIMEOUT_ERROR;

					tuner->TC90512->PSK_Write_pskmsrst();
					break;
				}
				case ISDB_T:
				{
					bool catv;
					uint number, frequency;
					GetTerrestrialFrequency(channel, &catv, &number, &frequency);

					tuner->TDA6651->Write(7*frequency+1+offset, TDA6651::Mode_Normal);

					bool locked = false;
					for (uint i=0; i<50; i++) {
						tuner->TDA6651->Read(&locked);
						if (locked) break;
						OS::Thread::Sleep((1)*1000*1000);
					}
					if (locked == false) return STATUS_PLL_LOCK_TIMEOUT_ERROR;

					tuner->TC90512->OFDM_Write_imsrst();
					break;
				}
			}

			tuner->FrequencyChannel[isdb] = channel;
			tuner->FrequencyOffset [isdb] = offset;

			return status;
		}

		virtual status GetFrequency(uint tunerIndex, ISDB isdb, uint *channel, int *offset) const
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			if ((channel == NULL) || (offset == NULL)) return STATUS_INVALID_PARAM_ERROR;

			const Tuner *tuner = &mTuner[tunerIndex];
			*channel = tuner->FrequencyChannel[isdb];
			*offset  = tuner->FrequencyOffset [isdb];

			return status;
		}

		virtual status SetIdS(uint tunerIndex, uint id)
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];
			tuner->TC90512->PSK_Write_Id(static_cast<ushort>(id));

			return status;
		}

		virtual status GetIdS(uint tunerIndex, uint *id)
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			if (id == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			ushort shortId;
			tuner->TC90512->PSK_Read_Id(&shortId);
			
			*id = shortId;

			return STATUS_OK;
		}

		virtual status GetCnAgc(uint tunerIndex, ISDB isdb, uint *cn100, uint *currentAgc, uint *maxAgc)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90512->PSK_Read_CN(cn100);

					byte byteAgc;
					tuner->TC90512->PSK_Read_AGC(&byteAgc);
					_ASSERTE(currentAgc); *currentAgc = byteAgc;
					_ASSERTE(maxAgc); *maxAgc = 127;
					break;
				}
				case ISDB_T:
				{
					tuner->TC90512->OFDM_Read_cndat(cn100);

					byte byteAgc;
					tuner->TC90512->OFDM_Read_ifagc_dt(&byteAgc);
					_ASSERTE(currentAgc); *currentAgc = byteAgc;
					_ASSERTE(maxAgc); *maxAgc = 255;
					break;
				}
			}

			return status;
		}

		virtual status GetFrequencyOffset(uint tunerIndex, ISDB isdb, int *clock, int *carrier)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90512->PSK_Read_clkfrq(clock);
					tuner->TC90512->PSK_Read_afcfrq(carrier);
					break;
				}
				case ISDB_T:
				{
					tuner->TC90512->OFDM_Read_clkafc_dt(clock);
					tuner->TC90512->OFDM_Read_carafc_dt(carrier);
					break;
				}
			}

			return status;
		}

		virtual status GetCorrectedErrorRate(uint tunerIndex, ISDB isdb, LayerIndex layerIndex, ErrorRate *errorRate)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90512->PSK_Read_perr_pecy(layerIndex, &errorRate->Numerator, &errorRate->Denominator);
					break;
				}
				case ISDB_T:
				{
					tuner->TC90512->OFDM_Read_perr_pecy(layerIndex, &errorRate->Numerator, &errorRate->Denominator);
					break;
				}
			}

			return status;
		}

		virtual status ResetCorrectedErrorCount(uint tunerIndex, ISDB isdb)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90512->PSK_Write_jperst();
					break;
				}
				case ISDB_T:
				{
					tuner->TC90512->OFDM_Write_perst();
					break;
				}
			}

			return STATUS_OK;
		}

		virtual status GetErrorCount(uint tunerIndex, ISDB isdb, uint *count)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			if (count == NULL) return STATUS_INVALID_PARAM_ERROR;

			mXC3S->Counter()->GetError(tunerIndex*2+isdb, count);

			return status;
		}

		virtual status GetTmccT(uint tunerIndex, TmccT *tmcc)
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			bool b = false;
			while (true) {
				bool retryov;
				bool tmunvld;
				bool fulock;
				tuner->TC90512->OFDM_Read_retryov_tmunvld_fulock(&retryov, &tmunvld, &fulock);
				if (!fulock) {
					b = true;
					break;
				} else {
					if (retryov) {
						break;
					}
				}
				OS::Thread::Sleep((1)*1000*1000);
			}

			if (b) {
				tuner->TC90512->OFDM_Read_TMCC(tmcc);

//				tuner->TC90512->OFDM_Read_eqqth(&tmcc->DriverInternalUse);
				//::wprintf(L"[%d]", tmcc->DriverInternalUse);
			}

			return b ? STATUS_OK : STATUS_GENERAL_ERROR;
		}

		virtual status GetTmccS(uint tunerIndex, TmccS *tmcc)
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];
			bool b = tuner->TC90512->PSK_Read_TMCC(tmcc);

			return b ? STATUS_OK : STATUS_GENERAL_ERROR;
		}

		virtual status GetLayerS(uint tunerIndex, LayerS *layer)
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];
			tuner->TC90512->PSK_Read_Layer(layer);

			return status;
		}

		virtual status GetLockedT(uint tunerIndex, bool locked[3])
		{
			status status;

			status = Status(tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			bool rlocka, rlockb, rlockc;
			tuner->TC90512->OFDM_Read_rlock(&rlocka, &rlockb, &rlockc);

			_ASSERTE(locked);
			locked[0] = rlocka;
			locked[1] = rlockb;
			locked[2] = rlockc;

			if (rlockb) {
				uint ffsize, gdleng, seqen;
				tuner->TC90512->OFDM_Read_ffsize_gdleng_seqen(&ffsize, &gdleng, &seqen);
				_ASSERTE(8 <= seqen);
			}

			return status;
		}

		virtual status SetLayerEnable(uint tunerIndex, ISDB isdb, LayerMask layerMask)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					bool null_h = !(layerMask & LAYER_MASK_H);
					tuner->TC90512->PSK_Write_oponff(null_h);
					break;
				}
				case ISDB_T:
				{
					tuner->TC90512->OFDM_Write_laysel(static_cast<byte>(layerMask));
					break;
				}
			}

			tuner->LayerMask[isdb] = layerMask;

			return status;
		}

		virtual status GetLayerEnable(uint tunerIndex, ISDB isdb, LayerMask *layerMask) const
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			const Tuner *tuner = &mTuner[tunerIndex];

			if (layerMask == NULL) return PT::STATUS_INVALID_PARAM_ERROR;

			*layerMask = tuner->LayerMask[isdb];

			return PT::STATUS_OK;
		}

		// DMA

		virtual status SetBufferInfo(const BufferInfo *bufferInfo)
		{
			status status;

			status = Status();
			if (status) return status;

			if (mTransferEnable) return STATUS_BUFFER_IS_IN_USE_ERROR;

			status = mBuffer->SetBufferInfo(bufferInfo);

			return status;
		}

		virtual status GetBufferInfo(BufferInfo *bufferInfo) const
		{
			status status;

			status = Status();
			if (status) return status;
			
			status = mBuffer->GetBufferInfo(bufferInfo);

			return STATUS_OK;
		}

		virtual status GetBufferPtr(uint index, void **ptr) const
		{
			status status;

			status = Status();
			if (status) return status;

			status = mBuffer->GetBufferPtr(index, ptr);

			return status;
		}

		virtual status SetStreamEnable(uint tunerIndex, ISDB isdb, bool enable)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			status = mXC3S->Initiator()->SetStreamEnable(tunerIndex, isdb, enable);

			return status;
		}

		virtual status GetStreamEnable(uint tunerIndex, ISDB isdb, bool *enable) const
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			status = mXC3S->Initiator()->GetStreamEnable(tunerIndex, isdb, enable);

			return status;
		}

		virtual status SetStreamGray(uint tunerIndex, ISDB isdb, uint gray)
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			status = mXC3S->Initiator()->SetStreamGray(tunerIndex, isdb, gray);

			return status;
		}

		virtual status GetStreamGray(uint tunerIndex, ISDB isdb, uint *gray) const
		{
			status status;

			status = Status(tunerIndex, isdb);
			if (status) return status;

			status = mXC3S->Initiator()->GetStreamGray(tunerIndex, isdb, gray);

			return status;
		}

		virtual status ResetTransferCounter()
		{
			status status;

			status = Status();
			if (status) return status;

			mXC3S->Initiator()->Reset();

			return status;
		}

		virtual status IncrementTransferCounter()
		{
			status status;

			status = Status();
			if (status) return status;

			mXC3S->Initiator()->Increment();

			return status;
		}

		virtual status SetTransferEnable(bool enable)
		{
			status status;

			status = Status();
			if (status) return status;

			if (enable == false) {
				mXC3S->Initiator()->End();

				for (uint i=0; i<10; i++) {
					XC3S::Status::Info info;
					mXC3S->Status()->GetInfo(&info);
					if (info.InitiatorBusy == false) break;
					OS::Thread::Sleep((1)*1000*1000);
				}

				mTransferEnable = false;
				return status;
			}

			BufferInfo bufferInfo;
			status = mBuffer->GetBufferInfo(&bufferInfo);
			if (status) return status;

			if (bufferInfo.VirtualSize == 0) return STATUS_BUFFER_IS_NOT_ALLOCATED_ERROR;

//			mBuffer->Link();

			uint startAddress;
			status = mBuffer->GetStartAddress(&startAddress);
			if (status) return status;

			mXC3S->Initiator()->Begin(startAddress);

			mTransferEnable = true;

			return status;
		}

		virtual status GetTransferEnable(bool *enable) const
		{
			status status;

			status = Status();
			if (status) return status;

			if (enable == NULL) return STATUS_INVALID_PARAM_ERROR;

			*enable = mTransferEnable;

			return status;
		}

		virtual status GetTransferInfo(TransferInfo *transferInfo)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mXC3S->Status()->GetTransferInfo(transferInfo);
			if (status) return status;

			return status;
		}

	protected:
		void (*const mBusFunction)(Device_ *, void *);
		void *const mBusArgument;
		PCI::_ *const mPCI;
		Memory::_ *const mMemory;
		const Bus::DeviceInfo mDeviceInfo;

		PCI::Device *mDevice;
		XC3S::_ *mXC3S;
		AT45DB041D *mAT45DB041D;

		Device2_ *mDevice2;
		Buffer *mBuffer;
		bool mTransferEnable;

		struct Tuner {
			TC90512 *TC90512;
			TDA6651 *TDA6651;
			QM1D1B0004 *QM1D1B0004;

			bool Sleep[2];
			uint FrequencyChannel[2];
			int  FrequencyOffset[2];
			LayerMask LayerMask[2];
		};

		Tuner mTuner[2];

		status SetConfig(uint *address_)
		{
			XC3S::Config	*config = mXC3S->Config();
			_ASSERTE(config);

			uint address;
			config->GetAddress(&address);
			if (address) {
				_ASSERTE(address_);
				*address_ = address;
			} else {
				#if defined(_DEBUG)
				_ASSERTE(address_);
				address = *address_;
				config->SetAddress(address);
				#else
				return STATUS_PCI_BASE_ADDRESS_ERROR;
				#endif
			}

			XC3S::Config::Enable enable;
			enable.Memory = true;
			enable.Master = true;

			config->SetEnable(&enable);

			return STATUS_OK;
		}

		void GetSatelliteFrequency(uint channel, bool *bs, uint *number, uint *frequency) const
		{
			_ASSERTE(bs);
			_ASSERTE(number);
			_ASSERTE(frequency);

			if (channel < 12) {
				*bs = true;
				*number = 1 + 2*channel;
				*frequency = 104948 + 3836*channel;
			} else if (channel < 24) {
				channel -= 12;

				*bs = false;
				*number = 2 + 2*channel;
				*frequency = 161300 + 4000*channel;
			} else {
				channel -= 24;

				*bs = false;
				*number = 1 + 2*channel;
				*frequency = 159300 + 4000*channel;
			}
		}

		void GetTerrestrialFrequency(uint channel, bool *catv, uint *number, uint *frequency) const
		{
			int frequencyOffset = 0;
			if (12 <= channel) frequencyOffset += 2;
			if (17 <= channel) frequencyOffset -= 2;
			if (63 <= channel) frequencyOffset += 2;
			*frequency = 93 + channel * 6 + frequencyOffset;

			static const byte TABLE[][3] = {
				{   2,  0,   3 },
				{  12,  1,  22 },
				{  21,  0,  12 },
				{  62,  1,  63 },
				{ 112,  0,  62 }
			};

			for (uint i=0; i<sizeof(TABLE)/sizeof(*TABLE); i++) {
				if (channel <= TABLE[i][0]) {
					_ASSERTE(catv); *catv = TABLE[i][1] ? true : false;
					_ASSERTE(number); *number = channel + TABLE[i][2] - TABLE[i][0];
					break;
				}
			}
		}

		status Status() const
		{
			if (mDevice == NULL) return STATUS_DEVICE_IS_NOT_OPEN_ERROR;
			return STATUS_OK;
		}

		status Status(uint tunerIndex) const
		{
			if (mDevice == NULL) return STATUS_DEVICE_IS_NOT_OPEN_ERROR;
			if (2 <= tunerIndex) return STATUS_INVALID_PARAM_ERROR;
			return STATUS_OK;
		}

		status Status(uint tunerIndex, ISDB isdb) const
		{
			if (mDevice == NULL) return STATUS_DEVICE_IS_NOT_OPEN_ERROR;
			if ((2 <= tunerIndex) || (ISDB_COUNT <= isdb)) return STATUS_INVALID_PARAM_ERROR;
			return STATUS_OK;
		}
	};
}
}

#endif
